@@ -1,5 +1,5 @@
 from __future__ import absolute_import, print_function, division
-from pony.py23compat import PY37, PYPY, PY38, PY39, PY310, PY311
+from pony.py23compat import PY36, PY37, PY38, PY39, PY310, PY311, PYPY
 
 import sys, types, inspect
 from opcode import opname as opnames, HAVE_ARGUMENT, EXTENDED_ARG, cmp_op
@@ -31,11 +31,12 @@ def decompile(x):
     elif t is types.GeneratorType:
         codeobject = x.gi_frame.f_code
     elif t is types.FunctionType:
+        x = inspect.unwrap(x)
         codeobject = x.__code__
         if x.__closure__:
             cells = dict(zip(codeobject.co_freevars, x.__closure__))
     else:
-        throw(TypeError)
+        throw(TypeError("Can't decompile %r" % t))
     key = get_codeobject_id(codeobject)
     result = ast_cache.get(key)
     if result is None:
@@ -164,7 +165,6 @@ class Decompiler(object):
         if decompiler.stack:
             throw(DecompileError, 'Compiled code should represent a single expression')
     def get_instructions(decompiler):
-        PY36 = sys.version_info >= (3, 6)
         before_yield = True
         code = decompiler.code
         co_code = code.co_code
@@ -173,7 +173,6 @@ class Decompiler(object):
         while decompiler.pos < decompiler.end:
             i = decompiler.pos
             op = code.co_code[i]
-            opname = opnames[op].replace('+', '_')
             if PY36:
                 extended_arg = 0
                 oparg = code.co_code[i+1]
@@ -194,6 +193,8 @@ class Decompiler(object):
                         i += 1
                         oparg = co_code[i] + co_code[i + 1] * 256 + oparg * 65536
                         i += 2
+
+            opname = opnames[op].replace('+', '_')
             if op >= HAVE_ARGUMENT:
                 if op in hasconst:
                     arg = [code.co_consts[oparg]]
@@ -705,9 +706,15 @@ class Decompiler(object):
     def IS_OP(decompiler, invert):
         return decompiler.COMPARE_OP('is not' if invert else 'is')
 
-    def LIST_APPEND(decompiler, offset=None):
-        throw(InvalidQuery('Use generator expression (... for ... in ...) '
-                           'instead of list comprehension [... for ... in ...] inside query'))
+    def LIST_APPEND(decompiler, offset):
+        tos = decompiler.stack.pop()
+        list_node = decompiler.stack[-offset]
+        if isinstance(list_node, ast.comprehension):
+            throw(InvalidQuery('Use generator expression (... for ... in ...) '
+                               'instead of list comprehension [... for ... in ...] inside query'))
+
+        assert isinstance(list_node, ast.List), list_node
+        list_node.elts.append(tos)
 
     def LIST_EXTEND(decompiler, offset):
         if offset != 1:
@@ -724,6 +731,12 @@ class Decompiler(object):
         lst.elts.extend(values)
         return lst
 
+    def LIST_TO_TUPLE(decompiler):
+        tos = decompiler.stack.pop()
+        if not isinstance(tos, ast.List):
+            throw(InvalidQuery, "Translation error, please contact developers: list expected, got: %r" % tos)
+        return ast.Tuple(tos.elts, ast.Load())
+
     def LOAD_ATTR(decompiler, attr_name):
         return ast.Attribute(decompiler.stack.pop(), attr_name, ast.Load())
 
